#include "../Framework/Framework.h"
#include "../Framework/Util/T3dLoader.h"

struct TexVertex {
	Vertex vertex;
	float s, t;
//	Vertex sVec, tVec, normal;
};

#define DEF D3DDECLMETHOD_DEFAULT

D3DVERTEXELEMENT9 texVertexFormat[] = {
	{ 0, 0,                                      D3DDECLTYPE_FLOAT3, DEF, D3DDECLUSAGE_POSITION, 0},
	{ 0, 1 * sizeof(Vertex),                     D3DDECLTYPE_FLOAT2, DEF, D3DDECLUSAGE_TEXCOORD, 0},
//	{ 0, 1 * sizeof(Vertex) + 2 * sizeof(float), D3DDECLTYPE_FLOAT3, DEF, D3DDECLUSAGE_TEXCOORD, 1},
//	{ 0, 2 * sizeof(Vertex) + 2 * sizeof(float), D3DDECLTYPE_FLOAT3, DEF, D3DDECLUSAGE_TEXCOORD, 2},
//	{ 0, 3 * sizeof(Vertex) + 2 * sizeof(float), D3DDECLTYPE_FLOAT3, DEF, D3DDECLUSAGE_TEXCOORD, 3},
	D3DDECL_END()
};

struct State {
	Material *material;
	unsigned int indexIndex, nIndices;
};

struct FogVolume {
	FogVolume(){}
	FogVolume(const Vertex &pos, const float r){
		position = pos;
		radius = r;
	}

	Vertex position;
	float radius;
};

/* ---------------------------------------------------------- */

class MainApp : public Direct3DApp {
protected:
	TextureID tex0, tex1, tex2, tex3;

	ShaderID vertexShader, fragmentShader;	

    MaterialTable matTable;

	Set <State> states;
	FogVolume fogVolumes[5];

	TexVertex *vertices;
	unsigned int *indices;
	unsigned int nVertices, nIndices;

	LPDIRECT3DVERTEXBUFFER9 vertexBuffer;
	LPDIRECT3DINDEXBUFFER9 indexBuffer;
	LPDIRECT3DVERTEXDECLARATION9 vDecl;

public:
	bool checkBaseCapabilities();

	void resetCamera();
	void selectPixelFormat(PixelFormat &pixelFormat);

	bool init();
	bool load();
	bool unload();
	bool drawFrame();
};

bool MainApp::checkBaseCapabilities(){

	if (caps.PixelShaderVersion < D3DPS_VERSION(1,0) || caps.VertexShaderVersion < D3DVS_VERSION(2,0)){
		String error("This demo requires at least pixel shader 1.0 and vertex shader 2.0.\n"
					 "Your card and driver only supports pixel shader ");
		error.sprintf("%d.%d", (caps.PixelShaderVersion  >> 8) & 255, caps.PixelShaderVersion  & 255);
		error += " and vertex shader ";
		error.sprintf("%d.%d", (caps.VertexShaderVersion >> 8) & 255, caps.VertexShaderVersion & 255);
		error += "\n\nThis demo will now exit.";

		ALERT(error, "Error");
		return false;
	}

	return true;
}

void MainApp::resetCamera(){
	position = Vertex(450, -140, -120);
	Wx = -0.063f;
	Wy = 1.214f;
	Wz = 0;
}

void MainApp::selectPixelFormat(PixelFormat &pixelFormat){
	pixelFormat.redBits   = 8;
	pixelFormat.greenBits = 8;
	pixelFormat.blueBits  = 8;
	pixelFormat.alphaBits = 0;

	pixelFormat.depthBits = 16;
}

bool MainApp::init(){
	projection.loadProjectionMatrix(PI / 2, 0.75f, 2, 3000);
	return true;
}

bool polygonFilter(const int flags){
	return (flags & PF_Invisible) == 0 && (flags & PF_Translucent) == 0 && (flags >= 0);
}

int comp(class Polygon *const &elem0, class Polygon *const &elem1){
	return (elem0->getMaterial() - elem1->getMaterial());
}

bool MainApp::load(){
	vertexShader   = renderer->addVertexShader("vertexshader.vsh");
	fragmentShader = renderer->addFragmentShader(0, "fragmentshader.fsh");

	tex0 = renderer->addTexture("../Textures/Wood.dds");
	tex1 = renderer->addTexture("../Textures/FieldStone.dds");
	tex2 = renderer->addTexture("../Textures/wooddecoration1.dds");
	tex3 = renderer->addTexture("../Textures/brickfloor.dds");

	textTexture = renderer->addTexture("../Textures/font2.dds");

	Set <class Polygon *> polys;

	Material *mat;

	mat = new Material();
	mat->setTextures(tex0);
    matTable.addMaterialName("Adr", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex1);
    matTable.addMaterialName("bmFloor3", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex2);
    matTable.addMaterialName("bmdirtyt", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex2);
    matTable.addMaterialName("bmTechWall2", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex2);
    matTable.addMaterialName("bmetalw4copy", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex3);
    matTable.addMaterialName("sh_cels", 256, 256, mat, 0);

	mat = new Material();
	mat->setTextures(tex3);
    matTable.addMaterialName("bmetalbase", 256, 256, mat, 0);

	T3dLoader t3dLoader;
	t3dLoader.loadFromFile("fractal.t3d", polys, matTable, polyCreator, polygonFilter, Vertex(0,0,0));

	tesselate(polys, 8192, 96);
	polys.sort(comp);

	unsigned int i, j, len = polys.getSize();
	
	nVertices = 0;
	for (i = 0; i < len; i++){
		nVertices += polys[i]->getnVertices();
	}
	nIndices = 3 * nVertices - 6 * len;

	vertices = new TexVertex[nVertices];
	indices = new unsigned int[nIndices];

	unsigned int *iDest = indices;
	unsigned int fvai, vai = 0;
	unsigned int last = 0, ni = 0;

	mat = polys[0]->getMaterial();
	i = 0;
	do {
		if (i == len || mat != polys[i]->getMaterial()){
			State s;
			s.material = mat;
			s.indexIndex = last;
			s.nIndices = ni - last;
			states.add(s);

			if (i == len) break;

			last = ni;
			mat = polys[i]->getMaterial();
		}

		fvai = vai;
		for (j = 0; j < polys[i]->getnVertices(); j++){
			if (j > 2){
				*iDest++ = fvai;
				*iDest++ = vai - 1;
			}
			*iDest++ = vai;

			vertices[vai].vertex = polys[i]->getVertex(j);
			vertices[vai].s = (polys[i]->getVertex(j) - polys[i]->getOrigin()) * (polys[i]->getS() * polys[i]->getScaleS());
			vertices[vai].t = (polys[i]->getVertex(j) - polys[i]->getOrigin()) * (polys[i]->getT() * polys[i]->getScaleT());
			//vertices[vai].sVec = polys[i]->getS();
			//vertices[vai].tVec = polys[i]->getT();
			//vertices[vai].normal = polys[i]->getPlane().normal;

			vai++;
		}
		ni += 3 * (polys[i]->getnVertices() - 2);
		
		delete polys[i];

		i++;
	} while (true);

	void *dest;
	dev->CreateVertexBuffer(nVertices * sizeof(TexVertex), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &vertexBuffer, NULL);
	vertexBuffer->Lock(0, 0, &dest, 0);
	memcpy(dest, vertices, nVertices * sizeof(TexVertex));
	vertexBuffer->Unlock();

	dev->CreateIndexBuffer(nIndices * sizeof(unsigned int), D3DUSAGE_WRITEONLY, D3DFMT_INDEX32, D3DPOOL_MANAGED, &indexBuffer, NULL);
	indexBuffer->Lock(0, 0, &dest, 0);
	memcpy(dest, indices, nIndices * sizeof(unsigned int));
	indexBuffer->Unlock();

	dev->CreateVertexDeclaration(texVertexFormat, &vDecl);

	fogVolumes[0] = FogVolume(Vertex(0, -50,  450), sqrf(180));
	fogVolumes[1] = FogVolume(Vertex(0, -50, -450), sqrf(180));
	fogVolumes[2] = FogVolume(Vertex(-890, -580, 0), sqrf(200));
	fogVolumes[3] = FogVolume(Vertex( 890, -580, 0), sqrf(200));

	return true;
}

bool MainApp::unload(){
	vDecl->Release();

	delete indices;
	delete vertices;

	indexBuffer->Release();
	vertexBuffer->Release();

	matTable.clear();
	states.clear();		

	return true;
}

bool MainApp::drawFrame(){
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0,0,0,0), 1.0f, 0);

	modelView.loadRotateZXY(-Wx, -Wy, -Wz);
	modelView.translate(-position);
	
	dev->SetTransform(D3DTS_VIEW,       (D3DMATRIX *) (const float *) modelView);
	dev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX *) (const float *) projection);


	fogVolumes[4].position = Vertex(0, 225, 0);
	fogVolumes[4].radius = sqrf(80 * sqrf((sinf(currTime) + 1)));



	dev->SetVertexShaderConstantF(0, transpose(projection * modelView), 4);

	float pos[] = { position.x, position.y, position.z, 0 };

	dev->SetVertexShaderConstantF(4, pos, 1);
	dev->SetVertexShaderConstantF(6, (float *) fogVolumes, sizeof(fogVolumes) / sizeof(FogVolume));

	int loop[] = { sizeof(fogVolumes) / sizeof(FogVolume), 6, 1, 0 };
	dev->SetVertexShaderConstantI(0, loop, 1);

	dev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	dev->SetVertexDeclaration(vDecl);
	dev->SetIndices(indexBuffer);
	dev->SetStreamSource(0, vertexBuffer, 0, sizeof(TexVertex));

	for (unsigned int i = 0; i < states.getSize(); i++){
		renderer->setTextures(states[i].material->textures[0]);
		renderer->setVertexShader(vertexShader);
		renderer->setFragmentShader(fragmentShader);
		renderer->apply();

		dev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, nVertices, states[i].indexIndex, states[i].nIndices / 3);
	}


	return true;
}

DECLARE_MAIN_APP(MainApp)
